""" spec2nii module for writing output NIFTI images
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford 
"""

import nibabel as nib
import os.path as op
import numpy as np

def writeNii(filename,outdir,data,positionInfo,dwelltime):
    """ Write spectroscopy data as NIFTI
    
    Args:
        filename (str): Output file name less extension
        outdir (str): Output directory
        data (np.ndarray): Time domain data (excluding spatial dimensions)
        positionInfo (NIFTIOrient obj): Orientation information.
        dwelltime (float): Dwelltime in seconds

    """
    # Form full path
    fullfilepath = op.join(outdir,filename+'.nii.gz')

    # Create new nifti image
    newobj = nib.nifti2.Nifti2Image(data,positionInfo.Q44)

    # Write new header
    pixDim = newobj.header['pixdim']
    pixDim[4] = dwelltime
    newobj.header['pixdim'] = pixDim
    
    # From nii obj and write    
    nib.save(newobj,fullfilepath)

    return None


""" spec2nii - tool for conversion of various MRS data formats to NIFTI format.

spec2nii converts the following formats to NIFTI files.
Supporting SVS:
    Siemens "Twix" .dat format
    Siemens DICOM
    Philips SPAR/SDAT files
    GE p-files
    LCModel RAW
    jMRUI text
    Plain text

Supporting CSI/MRSI:
    

This module contains the main class to be called as a script (through the main function).

Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford 
"""

import argparse
import sys
import numpy as np
from spec2nii.writeNii import writeNii
from spec2nii.writeJSON import writeJSON
import os.path as op
from pathlib import Path
from spec2nii.dcm2niiOrientation.orientationFuncs import nifti_dicom2mat
from spec2nii.nifti_orientation import NIFTIOrient
# There are case specific imports below


class spec2nii:
    def __init__(self):
        parser = argparse.ArgumentParser(description='Convert raw spectroscopy data to NIfTI format.')      

        subparsers = parser.add_subparsers(title='subcommands',
                                            description='File types supported')

        # Handle twix subcommand
        parser_twix = subparsers.add_parser('twix', help='Convert from Siemens .dat twix format.')
        parser_twix.add_argument('file',help='file to convert',type=str)
        parser_twix.add_argument('-j','--json',help='Create json sidecar.',action='store_true')
        group = parser_twix.add_mutually_exclusive_group(required=True)
        group.add_argument("-v", "--view", help="View contents of twix file, no files converted",action='store_true')
        group.add_argument('-e','--evalinfo', type=str, help='evalInfo flag to convert')
        parser_twix.add_argument("-f", "--fileout", type=str,help="Output file base name (default = input file name)")
        parser_twix.add_argument("-o", "--outdir", type=str,help="Output location (default = .)",default='.')
        parser_twix.add_argument("-m", "--multiraid", type=int,help="Select multiraid file to load (default = 2 i.e. 2nd file)",default=2)      
        parser_twix.add_argument("-q", "--quiet",help="Suppress text output",action='store_true')       
        parser_twix.set_defaults(func=self.twix)

        # Handle dicom subcommand
        parser_dicom = subparsers.add_parser('dicom', help='Convert from DICOM format.')
        parser_dicom.add_argument('file',help='file or directory to convert',type=str)
        parser_dicom.add_argument("-f", "--fileout", type=str,help="Output file base name (default = DCM SeriesDescription tag)")
        parser_dicom.add_argument("-o", "--outdir", type=str,help="Output location (default = .)",default='.')
        parser_dicom.add_argument('-j','--json',help='Create json sidecar.',action='store_true')
        parser_dicom.set_defaults(func=self.dicom)

        # Handle philips subcommand
        parser_philips = subparsers.add_parser('philips', help='Convert from Philips spar/sdat format.')
        parser_philips.add_argument('sdat',help='SDAT file',type=str)
        parser_philips.add_argument('spar',help='SPAR file',type=str)
        parser_philips.add_argument("-f", "--fileout", type=str,help="Output file base name (default = SDAT/SPAR name)")
        parser_philips.add_argument("-o", "--outdir", type=str,help="Output location (default = .)",default='.')
        parser_philips.add_argument('-j','--json',help='Create json sidecar.',action='store_true')
        parser_philips.set_defaults(func=self.philips)

        # Handle GE subcommand
        parser_ge = subparsers.add_parser('ge', help='Convert from GE p-file format.')
        parser_ge.add_argument('file',help='file to convert',type=str)
        parser_ge.add_argument("-f", "--fileout", type=str,help="Output file name (default = input name)")
        parser_ge.add_argument("-o", "--outdir", type=str,help="Output location (default = .)",default='.')
        parser_ge.add_argument('-j','--json',help='Create json sidecar.',action='store_true')
        parser_ge.set_defaults(func=self.ge)

        # Handle ismrmrd subcommand
        parser_ismrmrd = subparsers.add_parser('ismrmrd', help='Convert from ismrmrd format.')
        parser_ismrmrd.add_argument('file',help='file to convert',type=str)
        parser_ismrmrd.add_argument('a', type=str, help='placeholder')
        parser_ismrmrd.set_defaults(func=self.ismrmrd)

        # Handle text subcommand
        parser_txt = subparsers.add_parser('text', help='Convert from plain text format.')
        parser_txt.add_argument('file',help='file to convert',type=str)
        parser_txt.add_argument('-j','--json',help='Create json sidecar.',action='store_true')
        parser_txt.add_argument("-i", "--imagingfreq", type=float,help="Imaging (central) frequency in MHz",required=True)
        parser_txt.add_argument("-b", "--bandwidth", type=float,help="Reciever bandwidth (sweepwidth) in Hz.",required=True)
        parser_txt.add_argument("-a", "--affine", type=str,help="NIfTI affine file",required=False,metavar='<file>')
        parser_txt.add_argument("-f", "--fileout", type=str,help="Output file base name (default = input file name)")
        parser_txt.add_argument("-o", "--outdir", type=str,help="Output location (default = .)",default='.')        
        parser_txt.set_defaults(func=self.text)

        parser_jmrui = subparsers.add_parser('jmrui', help='Convert from jMRUI text format.')
        parser_jmrui.add_argument('file',help='file to convert',type=str)
        parser_jmrui.add_argument('-j','--json',help='Create json sidecar.',action='store_true')
        parser_jmrui.add_argument("-a", "--affine", type=str,help="NIfTI affine file",required=False,metavar='<file>')
        parser_jmrui.add_argument("-f", "--fileout", type=str,help="Output file base name (default = input file name)")
        parser_jmrui.add_argument("-o", "--outdir", type=str,help="Output location (default = .)",default='.')        
        parser_jmrui.set_defaults(func=self.jmrui)

        parser_raw = subparsers.add_parser('raw', help='Convert from LCModel RAW text format.')
        parser_raw.add_argument('file',help='file to convert',type=str)
        parser_raw.add_argument('-j','--json',help='Create json sidecar.',action='store_true')
        parser_raw.add_argument("-a", "--affine", type=str,help="NIfTI affine file",required=False,metavar='<file>')
        parser_raw.add_argument("-f", "--fileout", type=str,help="Output file base name (default = input file name)")
        parser_raw.add_argument("-o", "--outdir", type=str,help="Output location (default = .)",default='.')        
        parser_raw.set_defaults(func=self.raw)

        parser.add_argument('--verbose',action='store_true')        

        if len(sys.argv)==1:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        args = parser.parse_args()        
 
        self.fileoutNames = []
        self.imageOut = []
        self.orientationInfoOut = []
        self.dwellTimes = []
        self.metaData = []

        self.outputDir = args.outdir

        args.func(args)

        if self.imageOut:
            # Write nifti files            
            Path(self.outputDir).mkdir(parents=True, exist_ok=True)

            for n,i,o,d in zip(self.fileoutNames,self.imageOut,self.orientationInfoOut,self.dwellTimes):
                writeNii(n,self.outputDir,i,o,d)
            
            if self.metaData and args.json:
                for n,m in zip(self.fileoutNames,self.metaData):
                    writeJSON(n,self.outputDir,m)

        elif hasattr(args, 'view') and not args.view:
            print('No files to write.')

    def twix(self,args):     
        # Call mapvbvd to load the twix file.
        from mapvbvd import mapVBVD
        from spec2nii.twixfunctions import twix2DCMOrientation,examineTwix,extractTwixMetadata
        twixObj = mapVBVD(args.file,quiet=args.quiet)

        if  args.view:
            examineTwix(twixObj,op.basename(args.file),args.multiraid)
            return
        
        if isinstance(twixObj,list):                       
            twixObj = twixObj[args.multiraid-1]

        if not args.quiet:    
            print(f"Converting twix file {args.file}.")
            print(f'Looking for evalinfo flag {args.evalinfo}.')
        dataKey = args.evalinfo

        # Set squeeze data
        twixObj[dataKey].squeeze = True
        squeezedData = twixObj[dataKey]['']
        if not args.quiet: 
            print(f'Found data of size {squeezedData.shape}.')

        # Orientation calculations
        #1) Calculate dicom like imageOrientationPatient,imagePositionPatient,pixelSpacing and slicethickness
        imageOrientationPatient,imagePositionPatient,pixelSpacing,slicethickness = twix2DCMOrientation(twixObj['hdr'],verbose=args.verbose)
        # print(imageOrientationPatient)
        # print(imagePositionPatient)
        # print(pixelSpacing)
        # print(slicethickness)
        # 2) in style of dcm2niix
        # a) calculate Q44
        xyzMM = np.append(pixelSpacing,slicethickness)
        Q44 = nifti_dicom2mat(imageOrientationPatient,imagePositionPatient,xyzMM,verbose=args.verbose)
        # b) calculate nifti quaternion parameters
        Q44[:2,:] *= -1
        # qb,qc,qd,qx,qy,qz,dx,dy,dz,qfac = nifti_mat44_to_quatern(Q44)
        # 3) place in data class for nifti orientation parameters  
        currNiftiOrientation = NIFTIOrient(Q44)

        # Extract dwellTime
        dwellTime = twixObj['hdr']['MeasYaps'][('sRXSPEC','alDwellTime','0')]/1E9

        if args.json:
            # Extract metadata
            meta = extractTwixMetadata(twixObj['hdr'])

        # Identify what those indicies are
        # If cha is one: loop over 3rd and higher dims and make 2D images
        # If cha isn't present one: loop over 2nd and higher dims and make 1D images
        # Don't write here, just fill up class property lists for later writing
        if args.fileout:
            mainStr = args.fileout
        else:
            mainStr = op.splitext(op.basename(args.file))[0]

        dims = twixObj[dataKey].sqzDims
        if dims[0] != 'Col':
            raise ValueError('Col is expected to be the first dimension in the Twix file, it is not.')# This is very unlikely to occur  but would cause complete failure.
        if 'Cha' in dims:
            if dims[1] != 'Cha':
                raise ValueError('If present Cha is expected to be the first dimension in the Twix file, it is not.') # This is very unlikely to occur but would cause complete failure.           
            if len(dims) ==2: # Only single file output, force singleton third dimensions
                squeezedData = squeezedData.reshape((squeezedData.shape+(1,)))
            # Loop over all the other dims from the third up
            for index in np.ndindex(squeezedData.shape[2:]):
                modIndex = (slice(None),slice(None),)+ index

                newshape = (1,1,1)+squeezedData[modIndex].shape
                data_out = squeezedData[modIndex].reshape(newshape)

                self.imageOut.append(data_out)
                self.orientationInfoOut.append(currNiftiOrientation)
                self.dwellTimes.append(dwellTime)
                if args.json:
                    self.metaData.append(meta)
                # Create strings
                for idx,ii in enumerate(index):
                    if len(dims) > 2:
                        indexStr = dims[2+idx]
                    else:
                        self.fileoutNames.append(f'{mainStr}')
                        continue

                    if idx==0:
                        self.fileoutNames.append(f'{mainStr}_{indexStr}{ii :03d}')
                    else:
                        self.fileoutNames[len(self.fileoutNames)-1] += f'_{indexStr}{ii :03d}'

        else: # Loop over all the other dims from the second up
            if len(dims) ==1: # Only single file output, force singleton second dimensions
                squeezedData = squeezedData.reshape((squeezedData.shape+(1,)))
            
            for index in np.ndindex(squeezedData.shape[1:]):
                modIndex = (slice(None),)+ index

                newshape = (1,1,1)+squeezedData[modIndex].shape
                data_out = squeezedData[modIndex].reshape(newshape)

                self.imageOut.append(data_out)
                self.orientationInfoOut.append(currNiftiOrientation)
                self.dwellTimes.append(dwellTime)
                if args.json:
                    self.metaData.append(meta)
                # Create strings
                for idx,ii in enumerate(index):
                    indexStr = dims[1+idx]
                    if idx==0:
                        self.fileoutNames.append(f'{mainStr}_{indexStr}{ii :03d}')
                    else:
                        self.fileoutNames[len(self.fileoutNames)-1] += f'_{indexStr}{ii :03d}'

    def dicom(self,args):
        path_in = Path(args.file)
        if path_in.is_dir():
            # Look for typical dicom file extensions
            files_in = sorted(path_in.glob('*.IMA')) + \
                       sorted(path_in.glob('*.ima')) + \
                       sorted(path_in.glob('*.dcm'))

            # If none found look for all files
            if len(files_in) ==0:
                files_in = sorted([x for x in path_in.iterdir() if x.is_file()])

            print(f'Found {len(files_in)} files.')
        else:
            print('Single file conversion.')
            files_in = [path_in]
        
        # DICOM specific imports
        import nibabel.nicom.dicomwrappers
        from spec2nii.dicomfunctions import svs_or_CSI,process_siemens_svs,process_siemens_csi

        for idx,fn in enumerate(files_in):
            print(f'Converting dicom file {fn}')
            
            img = nibabel.nicom.dicomwrappers.wrapper_from_file(fn)

            mrs_type = svs_or_CSI(img)

            if mrs_type == 'SVS':
                specDataCmplx,currNiftiOrientation,dwelltime,meta = process_siemens_svs(img,args)
            
                if args.fileout:
                    mainStr = args.fileout
                else: 
                    mainStr = img.dcm_data.SeriesDescription
                self.fileoutNames.append(f'{mainStr}_{idx :03d}')

                newshape = (1,1,1)+specDataCmplx.shape
                specDataCmplx = specDataCmplx.reshape(newshape)

                self.imageOut.append(specDataCmplx)
                self.orientationInfoOut.append(currNiftiOrientation)
                self.dwellTimes.append(dwelltime)
                if args.json:
                    self.metaData.append(meta)

            elif mrs_type == 'CSI':
                specDataCmplx,currNiftiOrientation,dwelltime,meta = process_siemens_csi(img,args)
            
                if args.fileout:
                    mainStr = args.fileout
                else: 
                    mainStr = img.dcm_data.SeriesDescription
                self.fileoutNames.append(f'{mainStr}_{idx :03d}')

                self.imageOut.append(specDataCmplx)
                self.orientationInfoOut.append(currNiftiOrientation)
                self.dwellTimes.append(dwelltime)
                if args.json:
                    self.metaData.append(meta)


    def ismrmrd(self,args):
        print(f'Ismrmrd not yet handled!')
        print('exiting')
        return None

    def philips(self,args):
        # philips specific imports
        from spec2nii.philips import read_sdat_spar_pair

        data, orientation, dwelltime, meta = read_sdat_spar_pair(args.sdat,args.spar)

        # name of output
        if args.fileout:
            mainStr = args.fileout
        elif op.splitext(op.basename(args.sdat))[0]==op.splitext(op.basename(args.spar))[0]:
            mainStr = op.splitext(op.basename(args.sdat))[0]
        else:
            mainStr = op.splitext(op.basename(args.sdat))[0]

        newshape = (1,1,1)+data.shape
        data = data.reshape(newshape)

        # Place in data output format
        self.imageOut.append(data)
        self.orientationInfoOut.append(orientation)
        self.dwellTimes.append(dwelltime)
        self.metaData.append(meta)
        self.fileoutNames.append(mainStr)

    def ge(self,args):
        # philips specific imports
        from spec2nii.GE import read_p_file

        data,ref_data, orientation, dwelltime, meta = read_p_file(args.file)

        # name of output
        if args.fileout:
            baseStr = args.fileout
        else:
            baseStr = op.splitext(op.basename(args.file))[0]

        # Place in data output format
        for idx,d in enumerate(data):
            d = d.T
            newshape = (1,1,1)+d.shape
            d = d.reshape(newshape)
            self.imageOut.append(d)
            self.orientationInfoOut.append(orientation)
            self.dwellTimes.append(dwelltime)
            self.metaData.append(meta)
            self.fileoutNames.append(baseStr+f'_frame{idx:03.0f}')
        if ref_data is not None:
            for idx,d in enumerate(ref_data):
                d = d.T
                newshape = (1,1,1)+d.shape
                d = d.reshape(newshape)
                self.imageOut.append(d)
                self.orientationInfoOut.append(orientation)
                self.dwellTimes.append(dwelltime)
                self.metaData.append(meta)
                self.fileoutNames.append(baseStr+f'_ref_frame{idx:03.0f}')

    def text(self,args):
        # Read text from file
        data = np.loadtxt(args.file)
        data = data[:,0] + 1j*data[:,1] 
        
        newshape = (1,1,1)+data.shape
        data = data.reshape(newshape)

        # Interpret required arguments (frequency and bandwidth)
        imagingfreq = args.imagingfreq
        dwelltime = 1.0/args.bandwidth
        metadict = {'ImagingFrequency':imagingfreq,'Dwelltime':dwelltime}
        
        # Read optional affine file
        if args.affine:
            affine = np.loadtxt(args.affine)
        else:
            affine = np.eye(4)
        # qb,qc,qd,qx,qy,qz,dx,dy,dz,qfac = nifti_mat44_to_quatern(affine)        
        currNiftiOrientation = NIFTIOrient(affine)

        # File names
        if args.fileout:
            mainStr = args.fileout
        else:
            base=op.basename(args.file)
            mainStr = op.splitext(base)[0]

        # Place in data output format
        self.imageOut.append(data)
        self.orientationInfoOut.append(currNiftiOrientation)
        self.dwellTimes.append(dwelltime)
        self.metaData.append(metadict)
        self.fileoutNames.append(mainStr)

    def jmrui(self,args):
        from fsl_mrs.utils.mrs_io import jmrui_io
        # Read data from file
        data,header = jmrui_io.readjMRUItxt(args.file)

        newshape = (1,1,1)+data.shape
        data = data.reshape(newshape)

        # meta
        dwelltime = header['dwelltime']
        metadict = {'ImagingFrequency':header['centralFrequency'],'Dwelltime':header['dwelltime']}
        
        # Read optional affine file
        if args.affine:
            affine = np.loadtxt(args.affine)
        else:
            affine = np.eye(4)
        # qb,qc,qd,qx,qy,qz,dx,dy,dz,qfac = nifti_mat44_to_quatern(affine)        
        currNiftiOrientation = NIFTIOrient(affine)

        # File names
        if args.fileout:
            mainStr = args.fileout
        else:
            base=op.basename(args.file)
            mainStr = op.splitext(base)[0]

        # Place in data output format
        self.imageOut.append(data)
        self.orientationInfoOut.append(currNiftiOrientation)
        self.dwellTimes.append(dwelltime)
        self.metaData.append(metadict)
        self.fileoutNames.append(mainStr)

    def raw(self,args):
        from fsl_mrs.utils.mrs_io import lcm_io
        # Read data from file
        data,header = lcm_io.readLCModelRaw(args.file,conjugate=True)
        
        newshape = (1,1,1)+data.shape
        data = data.reshape(newshape)

        # meta
        dwelltime = header['dwelltime']
        metadict = {'ImagingFrequency':header['centralFrequency'],'Dwelltime':header['dwelltime']}
        
        # Read optional affine file
        if args.affine:
            affine = np.loadtxt(args.affine)
        else:
            affine = np.eye(4)
        # qb,qc,qd,qx,qy,qz,dx,dy,dz,qfac = nifti_mat44_to_quatern(affine)        
        currNiftiOrientation = NIFTIOrient(affine)

        # File names
        if args.fileout:
            mainStr = args.fileout
        else:
            base=op.basename(args.file)
            mainStr = op.splitext(base)[0]

        # Place in data output format
        self.imageOut.append(data)
        self.orientationInfoOut.append(currNiftiOrientation)
        self.dwellTimes.append(dwelltime)
        self.metaData.append(metadict)
        self.fileoutNames.append(mainStr)

def main(*args):
    spec2nii(*args)
    return 0

# if __name__== "__main__":
#     spec2nii()


""" spec2nii module containing functions specific to interpreting Siemens DICOM
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford 
"""
import numpy as np
from spec2nii.dcm2niiOrientation.orientationFuncs import nifti_dicom2mat
from spec2nii.nifti_orientation import NIFTIOrient

def svs_or_CSI(img):
    rows = img.csa_header['tags']['Rows']['items'][0]
    cols = img.csa_header['tags']['Columns']['items'][0]
    slices = img.csa_header['tags']['NumberOfFrames']['items'][0]

    if np.prod([rows,cols,slices])>1.0:
        return 'CSI'
    else:
        return 'SVS'

def process_siemens_svs(img,args):
    specData = np.frombuffer(img.dcm_data[('7fe1', '1010')].value, dtype=np.single)
    specDataCmplx = specData[0::2]+1j*specData[1::2]    

    #1) Extract dicom parameters
    imageOrientationPatient = np.array(img.csa_header['tags']['ImageOrientationPatient']['items']).reshape(2,3)
    imagePositionPatient = img.csa_header['tags']['VoiPosition']['items']
    xyzMM = np.array([img.csa_header['tags']['VoiPhaseFoV']['items'][0],
                    img.csa_header['tags']['VoiReadoutFoV']['items'][0],
                img.csa_header['tags']['VoiThickness']['items'][0]])
    # 2) in style of dcm2niix
    # a) calculate Q44
    Q44 = nifti_dicom2mat(imageOrientationPatient,imagePositionPatient,xyzMM)
    # b) calculate nifti quaternion parameters
    Q44[:2,:] *= -1
    # qb,qc,qd,qx,qy,qz,dx,dy,dz,qfac = nifti_mat44_to_quatern(Q44)
    # 3) place in data class for nifti orientation parameters 
    currNiftiOrientation = NIFTIOrient(Q44)
    dwelltime = img.csa_header['tags']['RealDwellTime']['items'][0]*1E-9
    meta = extractDicomMetadata(img)

    return specDataCmplx,currNiftiOrientation,dwelltime,meta

def process_siemens_csi(img,args):
    specData = np.frombuffer(img.dcm_data[('7fe1', '1010')].value, dtype=np.single)
    specDataCmplx = specData[0::2]-1j*specData[1::2]

    rows = img.csa_header['tags']['Rows']['items'][0]
    cols = img.csa_header['tags']['Columns']['items'][0]
    slices = img.csa_header['tags']['NumberOfFrames']['items'][0]
    spectral_points = img.csa_header['tags']['DataPointColumns']['items'][0]    

    specDataCmplx = specDataCmplx.reshape((slices,rows,cols,spectral_points))
    specDataCmplx = np.moveaxis(specDataCmplx,(0,1,2),(2,1,0))

    #1) Extract dicom parameters
    imageOrientationPatient = np.array(img.csa_header['tags']['ImageOrientationPatient']['items']).reshape(2,3)
    imagePositionPatient = np.array(img.csa_header['tags']['ImagePositionPatient']['items'])
    xyzMM = np.array([img.csa_header['tags']['PixelSpacing']['items'][0],
                  img.csa_header['tags']['PixelSpacing']['items'][1],
                  img.csa_header['tags']['SliceThickness']['items'][0]])
    # 2) in style of dcm2niix
    # a) calculate Q44
    Q44 = nifti_dicom2mat(imageOrientationPatient,imagePositionPatient,xyzMM)
    # b) calculate nifti quaternion parameters
    Q44[:2,:] *= -1
    # qb,qc,qd,qx,qy,qz,dx,dy,dz,qfac = nifti_mat44_to_quatern(Q44)
    # 3) place in data class for nifti orientation parameters 
    currNiftiOrientation = NIFTIOrient(Q44)
    dwelltime = img.csa_header['tags']['RealDwellTime']['items'][0]*1E-9
    meta = extractDicomMetadata(img)
    
    return specDataCmplx,currNiftiOrientation,dwelltime,meta


def extractDicomMetadata(dcmdata):
    """ Extract information from the nibabel DICOM objhect to insert into the json sidecar.
    
    Args:
        dcmdata: nibabel.nicom image object
    Returns:
        metaDict (dict): Json sidecard output
    """

    metaDict = {}
    metaDict.update({'Modality':'MR'})
    metaDict.update({'Manufacturer':dcmdata.dcm_data.Manufacturer})
    metaDict.update({'ManufacturersModelName':dcmdata.dcm_data.ManufacturerModelName})
    metaDict.update({'DeviceSerialNumber':dcmdata.dcm_data.DeviceSerialNumber})
    metaDict.update({'StationName':dcmdata.dcm_data.StationName}) # Not an exact match but close
    metaDict.update({'SoftwareVersions':dcmdata.dcm_data.SoftwareVersions})
    metaDict.update({'MagneticFieldStrength':dcmdata.csa_header['tags']['MagneticFieldStrength']['items'][0]})
    if len(dcmdata.csa_header['tags']['ReceivingCoil']['items'])>0:
        metaDict.update({'ReceiveCoilName':dcmdata.csa_header['tags']['ReceivingCoil']['items'][0]})
    else:
         metaDict.update({'ReceiveCoilName':dcmdata.csa_header['tags']['ImaCoilString']['items'][0]})
    
    metaDict.update({'ScanningSequence':dcmdata.csa_header['tags']['ScanningSequence']['items'][0]})
    metaDict.update({'SequenceVariant':''})
    metaDict.update({'ScanOptions':''})
    metaDict.update({'SequenceName':dcmdata.csa_header['tags']['SequenceName']['items'][0]})
    metaDict.update({'PulseSequenceDetails':''})
    metaDict.update({'EchoTime':dcmdata.csa_header['tags']['EchoTime']['items'][0]*1E-3})
    try: 
        metaDict.update({'InversionTime':dcmdata.csa_header['tags']['InversionTime']['items'][0]})
    except:
        metaDict.update({'InversionTime':''})
    metaDict.update({'Dwelltime':dcmdata.csa_header['tags']['RealDwellTime']['items'][0]*1E-9})
    metaDict.update({'FlipAngle':dcmdata.csa_header['tags']['FlipAngle']['items'][0]})
    metaDict.update({'InstitutionName':dcmdata.dcm_data.InstitutionName})
    metaDict.update({'InstitutionAddress':dcmdata.dcm_data.InstitutionAddress})
    metaDict.update({'InstitutionalDepartmentName':''})
    metaDict.update({'TimeStamp':dcmdata.dcm_data.AcquisitionTime})
    metaDict.update({'Nucleus':dcmdata.csa_header['tags']['ImagedNucleus']['items'][0]})
    metaDict.update({'ImagingFrequency':dcmdata.csa_header['tags']['ImagingFrequency']['items'][0]})
    metaDict.update({'RepetitionTime':dcmdata.csa_header['tags']['RepetitionTime']['items'][0]/1E3})
    if dcmdata.csa_header['tags']['PatientOrientation']['items']:
        metaDict.update({'PatientPosition':dcmdata.csa_header['tags']['PatientOrientation']['items'][0]})
    metaDict.update({'ProtocolName':dcmdata.dcm_data.ProtocolName})
    metaDict.update({'ConversionSoftware':'spec2nii dicom'})
    return metaDict


""" spec2nii module containing functions specific to Siemens TWIX format
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford 
"""

import spec2nii.GSL.gslfunctions as GSL
import numpy as np

def twix2DCMOrientation(mapVBVDHdr,verbose=False):
    """ Convert twix orientation information to DICOM equivalent.
    
    Convert orientation to DICOM imageOrientationPatient, imagePositionPatient, 
    pixelSpacing and sliceThickness field values.

    Args:
        mapVBVDHdr (dict): Header info interpreted by pymapVBVD
        verbose (bool,optionl)
    Returns:
        imageOrientationPatient
        imagePositionPatient
        pixelSpacing
        sliceThickness

    """
    #Orientation information
    if ('sSpecPara','sVoI','sNormal','dSag') in mapVBVDHdr['MeasYaps']:
        NormaldSag = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','sNormal','dSag')]
    else: 
        NormaldSag = 0.0
    
    if ('sSpecPara','sVoI','sNormal','dCor') in mapVBVDHdr['MeasYaps']:
        NormaldCor = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','sNormal','dCor')]
    else :
        NormaldCor = 0.0

    if ('sSpecPara','sVoI','sNormal','dTra') in mapVBVDHdr['MeasYaps']:
        NormaldTra = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','sNormal','dTra')]
    else:
        NormaldTra = 0.0
    
    if ('sSpecPara','sVoI','dInPlaneRot') in mapVBVDHdr['MeasYaps']: 
        inplaneRotation = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','dInPlaneRot')]
    else:
        inplaneRotation = 0.0    

    TwixSliceNormal =np.array([NormaldSag,NormaldCor,NormaldTra],dtype = float)

    RoFoV = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','dReadoutFOV')]
    PeFoV = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','dPhaseFOV')]              

    dColVec_vector, dRowVec_vector = GSL.fGSLCalcPRS(TwixSliceNormal,inplaneRotation,verbose)

    imageOrientationPatient = np.stack((dRowVec_vector,dColVec_vector),axis = 0)
    sliceNormal = TwixSliceNormal

    columns = 1
    rows = 1
    slices= 1

    pixelSpacing = np.array([PeFoV, RoFoV]) #[RoFoV PeFoV];
    sliceThickness = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','dThickness')]

    # Position info
    if ('sSpecPara','sVoI','sPosition','dSag') in mapVBVDHdr['MeasYaps']:
        PosdSag = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','sPosition','dSag')]
    else: 
        PosdSag = 0.0
    
    if ('sSpecPara','sVoI','sPosition','dCor') in mapVBVDHdr['MeasYaps']:
        PosdCor = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','sPosition','dCor')]
    else:
        PosdCor = 0.0

    if ('sSpecPara','sVoI','sPosition','dTra') in mapVBVDHdr['MeasYaps']:
        PosdTra = mapVBVDHdr['MeasYaps'][('sSpecPara','sVoI','sPosition','dTra')]
    else:
        PosdTra = 0.0

    basePosition =np.array([PosdSag, PosdCor, PosdTra],dtype = float)
    imagePositionPatient = basePosition
    if verbose:
        print(f'imagePositionPatient is {imagePositionPatient.ravel()}')
        print(f'imageOrientationPatient is \n{imageOrientationPatient}')
        print(f'{imageOrientationPatient.ravel()}')
        print(f'pixelSpacing is {pixelSpacing}')
        
    return imageOrientationPatient,imagePositionPatient,pixelSpacing,sliceThickness

def examineTwix(twixObj,fileName,mraid):
    """ Print formated twix contents"""

    print(f'Contents of file: {fileName}')

    if isinstance(twixObj,list):
        print(f'Multiraid file, {len(twixObj)} files found.')
        print(f'Selecting file {mraid}. Use -m option to change.')                       
        twixObj = twixObj[mraid-1]

    evalInfoFlags = twixObj.keys()
    evalInfoFlags = [i for i in evalInfoFlags if i != 'hdr']

    print(f'The file contains these evalinfo flags with dimensions and sizes as follows:')
    for ev in evalInfoFlags:
        twixObj[ev].squeeze = True
        tmpSqzSize = twixObj[ev].sqzSize
        tmpSqzDims = ', '.join(twixObj[ev].sqzDims)
        print(f'{ev: <15}:\t{tmpSqzDims: <20}\t{tmpSqzSize}')

    #import pdb; pdb.set_trace()

def extractTwixMetadata(mapVBVDHdr):
    """ Extract information from the pymapVBVD header to insert into the json sidecar.
    
    Args:
        dcmdata (dict): Twix headers
    Returns:
        metaDict (dict): Json sidecard output
    """

    metaDict = {}
    metaDict.update({'Modality':'MR'})
    metaDict.update({'Manufacturer':mapVBVDHdr['Dicom'][('Manufacturer')]})
    metaDict.update({'ManufacturersModelName':mapVBVDHdr['Dicom'][('ManufacturersModelName')]})
    metaDict.update({'DeviceSerialNumber':_try_int(mapVBVDHdr['Dicom'][('DeviceSerialNumber')])})
    metaDict.update({'StationName':_try_int(mapVBVDHdr['Dicom'][('DeviceSerialNumber')])}) # Not an exact match but close
    metaDict.update({'SoftwareVersions':mapVBVDHdr['Dicom'][('SoftwareVersions')]})
    metaDict.update({'MagneticFieldStrength':mapVBVDHdr['Dicom']['flMagneticFieldStrength']})
    if ('sCoilSelectMeas', 'aRxCoilSelectData', '0', 'asList', '0', 'sCoilElementID', 'tCoilID') in mapVBVDHdr['MeasYaps']:
        metaDict.update({'ReceiveCoilName':mapVBVDHdr['MeasYaps'][('sCoilSelectMeas', 'aRxCoilSelectData', '0', 'asList', '0', 'sCoilElementID', 'tCoilID')]})
    elif ('asCoilSelectMeas', '0', 'asList', '0', 'sCoilElementID', 'tCoilID') in  mapVBVDHdr['MeasYaps']:
         metaDict.update({'ReceiveCoilName':mapVBVDHdr['MeasYaps'][('asCoilSelectMeas', '0', 'asList', '0', 'sCoilElementID', 'tCoilID')]})
    else:
        metaDict.update({'ReceiveCoilName':''})
    metaDict.update({'ScanningSequence':mapVBVDHdr['Dicom'][('tScanningSequence')]})
    metaDict.update({'SequenceVariant':mapVBVDHdr['Dicom'][('tSequenceVariant')]})
    metaDict.update({'ScanOptions':mapVBVDHdr['Dicom'][('tScanOptions')]})
    metaDict.update({'SequenceName':''}) # Can't find in twix headers
    metaDict.update({'PulseSequenceDetails':mapVBVDHdr['Config'][('SequenceFileName')]})
    metaDict.update({'EchoTime':mapVBVDHdr['Phoenix'][('alTE','0')]*1E-6})
    if ('InversionTime') in mapVBVDHdr['Meas']:
        metaDict.update({'InversionTime':mapVBVDHdr['Meas'][('TI_Time')]})
    else:
        metaDict.update({'InversionTime':''})    
    metaDict.update({'Dwelltime':mapVBVDHdr['Meas'][('DwellTimeSig')]*1e-9})
    metaDict.update({'FlipAngle':mapVBVDHdr['Meas'][('FlipAngle')]})
    metaDict.update({'InstitutionName':mapVBVDHdr['Dicom'][('InstitutionName')]})
    metaDict.update({'InstitutionAddress':mapVBVDHdr['Dicom'][('InstitutionAddress')]})
    metaDict.update({'InstitutionalDepartmentName':''})#Not in twix headers
    metaDict.update({'TimeStamp':''})#Not in twix headers
    metaDict.update({'Nucleus':mapVBVDHdr['Meas'][('ResonantNucleus')]})
    metaDict.update({'ImagingFrequency':mapVBVDHdr['Meas'][('Frequency')]/1E6})
    if ('TR_Time') in mapVBVDHdr['Meas']:
        metaDict.update({'RepetitionTime':mapVBVDHdr['Meas'][('TR_Time')]/1E6}) 
    else:
        metaDict.update({'RepetitionTime':mapVBVDHdr['Meas'][('TR')]/1E6})     
    metaDict.update({'PatientPosition':mapVBVDHdr['Meas'][('PatientPosition')]})
    metaDict.update({'ProtocolName':mapVBVDHdr['Dicom'][('tProtocolName')]})
    metaDict.update({'ConversionSoftware':'spec2nii'})
    return metaDict

def _try_int(value):

    try:
        return int(value)
    except:
        return value


import numpy as np

def nifti_dicom2mat(orient,patientPosition,xyzMM,verbose = False):
    # DICOM values to 4x4 nifti transformation matrix. As per https://github.com/rordenlab/dcm2niix/blob/7ce33ca5fa3bb2dd4e5410bd97bcf515c9e462d9/console/nifti1_io_core.cpp

    Q = np.zeros((3,3))

    # Q.m[0][0] = orient[1]; Q.m[0][1] = orient[2] ; Q.m[0][2] = orient[3] ; // load Q
    # Q.m[1][0] = orient[4]; Q.m[1][1] = orient[5] ; Q.m[1][2] = orient[6];
    Q[:2,:] = orient # Fill first two rows
    
    # normalize rows
    #import pdb; pdb.set_trace()
    normQ = np.linalg.norm(Q, ord=2, axis=1, keepdims=True)
    normQ[normQ==0.0] = 1.0
    Q /= normQ
    if verbose:
        np.set_printoptions(precision=2)
        print(f'Normalised ImageOrientationPatient in Q:\n {Q}')

    # row 3 is the cross product of rows 1 and 2
    Q[2,:] = np.cross(Q[0,:],Q[1,:])
    if verbose:
        print(f'After slice normal calculation. Q:\n {Q}')

    Q = Q.T
    if np.linalg.det(Q)<0.0:
        Q[:,2] = Q[:,2] *-1

    # next scale matrix
    # I think dcm2niix reversees the pixel spacing: https://github.com/rordenlab/dcm2niix/blob/485c387c93bbca3b29b93403dfde211c4bc39af6/console/nii_dicom.cpp#L5403
    xyzMM[1], xyzMM[0] = xyzMM[0], xyzMM[1]
    diagVox = np.diag(xyzMM)
    Q = Q@diagVox
    if verbose:
        print(f'After scaling. Q:\n {Q}')
        
    # Include translations
    Q44 = np.zeros((4,4))
    Q44[:3,:3] = Q
    Q44[:3,3] = patientPosition
    Q44[3,3] = 1.0
    if verbose:
        print(f'Final Q:\n {Q44}')
    return Q44

def nifti_mat44_to_quatern(R):
    #offset outputs are read  out of input matrix
    qx = R[0,3]
    qy = R[1,3]
    qz = R[2,3]

    #load 3x3 matrix into local variables */
    r11 = R[0,0]
    r12 = R[0,1]
    r13 = R[0,2]
    r21 = R[1,0]
    r22 = R[1,1]
    r23 = R[1,2]
    r31 = R[2,0]
    r32 = R[2,1]
    r33 = R[2,2]

    # compute lengths of each column; these determine grid spacings
    xd = np.sqrt( r11*r11 + r21*r21 + r31*r31 )
    yd = np.sqrt( r12*r12 + r22*r22 + r32*r32 )
    zd = np.sqrt( r13*r13 + r23*r23 + r33*r33 )
    
    #if a column length is zero, patch the trouble
    if xd == 0.0:
        r11 = 1.0
        r21 = 0.0
        r31 = 0.0
        xd = 1.0
    if yd == 0.0:
        r22 = 1.0
        r12 = 0.0
        r32 = 0.0
        yd = 1.0        
    if zd == 0.0:
        r33 = 1.0
        r13 = 0.0
        r23 = 0.0
        zd = 1.0

    # assign the output lengths */
    dx = xd
    dy = yd
    dz = zd
    
    #normalize the columns */
    r11 /= xd
    r21 /= xd
    r31 /= xd
    r12 /= yd
    r22 /= yd
    r32 /= yd
    r13 /= zd
    r23 /= zd
    r33 /= zd  

    zd = r11*r22*r33-r11*r32*r23-r21*r12*r33+r21*r32*r13+r31*r12*r23-r31*r22*r13 # should be -1 or 1
    if zd > 0:# proper
        qfac = 1.0
    else: #%improper ==> flip 3rd column
        qfac=-1.0
        r13 *= -1.0
        r23 *= -1.0
        r33 *= -1.0
    
    # now, compute quaternion parameters
    a = r11 + r22 + r33 + 1.0
    if a > 0.5 :# simplest case
        a = 0.5 * np.sqrt(a)
        b = 0.25 * (r32-r23) / a
        c = 0.25 * (r13-r31) / a
        d = 0.25 * (r21-r12) / a
    else: #trickier case
        xd = 1.0 + r11 - (r22+r33) # 4*b*b
        yd = 1.0 + r22 - (r11+r33) # 4*c*c
        zd = 1.0 + r33 - (r11+r22) # 4*d*d
        if xd > 1.0:
            b = 0.5 * np.sqrt(xd)
            c = 0.25* (r12+r21) / b
            d = 0.25* (r13+r31) / b
            a = 0.25* (r32-r23) / b
        elif yd > 1.0:
            c = 0.5 * np.sqrt(yd)
            b = 0.25* (r12+r21) / c
            d = 0.25* (r23+r32) / c
            a = 0.25* (r13-r31) / c
        else:
            d = 0.5 * np.sqrt(zd)
            b = 0.25* (r13+r31) / d
            c = 0.25* (r23+r32) / d
            a = 0.25* (r21-r12) / d
        
        if a < 0.0:
            b=-b
            c=-c
            d=-d        

    qb = b
    qc = c
    qd = d
    return qb,qc,qd,qx,qy,qz,dx,dy,dz,qfac



from ast import literal_eval
import numpy as np
from spec2nii.nifti_orientation import NIFTIOrient,calc_affine
import warnings
def read_sdat_spar_pair(sdat_file,spar_file):

    spar_params = read_spar(spar_file)
    data = read_sdat(sdat_file,
                    spar_params['samples'],
                    spar_params['rows'])

    # Dwelltime
    dwelltime = 1.0 / float(spar_params["sample_frequency"])

    # Meta - which will be written to the JSON side car requires
    # a few parameters. Set these first. Then set the ones from
    # the spar file
    cf = float(spar_params["synthesizer_frequency"])/1E6
    meta = {'ImagingFrequency':cf,'Dwelltime':dwelltime}   
    meta.update(spar_params)    

    # Orientation
    affine = _philips_orientation(spar_params)
    orientation = NIFTIOrient(affine)

    return data, orientation, dwelltime, meta

def read_spar(filename):
    parameter_dict = {}
    with open(filename,'r') as f:        
        for line in f:
            # ignore comments (!) and empty lines
            if line == "\n" or line.startswith("!"):
                continue
                
            # Handle
            key, value = map(str.strip, line.split(":", 1))
            try:
                val = literal_eval(value)
            except:
                if value == '':
                    val = None
                else:
                    val = value
            
            parameter_dict.update({key:val})
            
    return parameter_dict

def read_sdat(filename,samples,rows):
    with open(filename, 'rb') as f:
        raw = f.read()
    
    floats = _vax_to_ieee_single_float(raw)
    data_iter = iter(floats)
    complex_iter = (complex(r, i) for r, i in zip(data_iter, data_iter))
    raw_data = np.fromiter(complex_iter, "complex64")
    raw_data = np.reshape(raw_data, (rows, samples)).squeeze()
    
    return raw_data

def _philips_orientation(params):
    warnings.warn('Philips orientation not yet tested.')   
    
    angle_lr = params['lr_angulation']
    angle_ap = params['ap_angulation']
    angle_hf = params['cc_angulation']
    angles = [-angle_lr,-angle_ap,angle_hf]

    dim_lr = params['lr_size']
    dim_ap = params['ap_size']
    dim_hf = params['cc_size']
    dimensions = [dim_lr,dim_ap,dim_hf]

    shift_lr = params['lr_off_center']
    shift_ap = params['ap_off_center']
    shift_hf = params['cc_off_center']
    shift = [-shift_lr,-shift_ap,shift_hf]

    return calc_affine(angles,dimensions,shift)

# From VESPA - BSD license. Ask for license text from Brian.
def _vax_to_ieee_single_float(data):
    """Converts a float in Vax format to IEEE format.

    data should be a single string of chars that have been read in from 
    a binary file. These will be processed 4 at a time into float values.
    Thus the total number of byte/chars in the string should be divisible
    by 4.
    
    Based on VAX data organization in a byte file, we need to do a bunch of 
    bitwise operations to separate out the numbers that correspond to the
    sign, the exponent and the fraction portions of this floating point
    number
    
    role :      S        EEEEEEEE      FFFFFFF      FFFFFFFF      FFFFFFFF
    bits :      1        2      9      10                               32
    bytes :     byte2           byte1               byte4         byte3    
    
    """
    f = []
    nfloat = int(len(data) / 4)
    for i in range(nfloat):

        byte2 = data[0 + i*4]
        byte1 = data[1 + i*4]
        byte4 = data[2 + i*4]
        byte3 = data[3 + i*4]

        # hex 0x80 = binary mask 10000000
        # hex 0x7f = binary mask 01111111

        sign = (byte1 & 0x80) >> 7
        expon = ((byte1 & 0x7f) << 1) + ((byte2 & 0x80) >> 7)
        fract = ((byte2 & 0x7f) << 16) + (byte3 << 8) + byte4

        if sign == 0:
            sign_mult = 1.0
        else:
            sign_mult = -1.0

        if 0 < expon:
            # note 16777216.0 == 2^24
            val = sign_mult * (0.5 + (fract/16777216.0)) * pow(2.0, expon - 128.0)
            f.append(val)
        elif expon == 0 and sign == 0:
            f.append(0)
        else:
            f.append(0)
            # may want to raise an exception here ...

    return f


#!/usr/bin/env python

from setuptools import setup
import versioneer

with open('requirements.txt', 'rt') as f:
    install_requires = [l.strip() for l in f.readlines()]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='spec2nii',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Multi-format in vivo MR spectroscopy conversion to NIFTI',
      author='Will Clarke',
      author_email='william.clarke@ndcn.ox.ac.uk',
      url='https://github.com/wexeee/spec2nii',
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=['spec2nii','spec2nii.GSL','spec2nii.dcm2niiOrientation'],
      install_requires=install_requires,
      classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        ],
      python_requires='>=3.7', 
      entry_points={"console_scripts": [
            "spec2nii = spec2nii.spec2nii:main"]
            }         
     )




""" spec2nii module for writing output sidecar JSON files
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford 
"""
import json
import os.path as op
import numpy as np

def writeJSON(filename,outdir,meta_info):
    """ Write spectroscopy meta-data as json
    
    Args:
        filename (str): Output file name, less extension
        outdir (str): Output directory        
        meta_info (dict): Meta info to write to json.
    """
    # Form full path
    fullfilepath = op.join(outdir,filename+'.json')

    # Clean up dict
    for key in meta_info:
        if isinstance(meta_info[key],np.ndarray):
            meta_info[key] = meta_info[key].tolist()

        # print(f'{key} is type: {type(meta_info[key])}.')

    with open(fullfilepath, 'w') as fp:
        json.dump(meta_info, fp)
    




# spec2nii, functions to interpret GE P files.
# Very hacky interpretation of GE p files focused on use for SVS MRS
# Cobbled together from three packages: Spant, Gannet and pfile-tools
# I have no documentation for or experience with GE data, so most of this is trial & error!
#
# Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
#
# Copyright (C) 2020 University of Oxford 

import numpy as np
import warnings
from spec2nii.nifti_orientation import NIFTIOrient,calc_affine

def read_p_file(filename):

    hdr = _read_hdr(filename)

    if (hdr['xcsi']*hdr['ycsi']*hdr['zcsi']) > 1:
        raise ValueError(f"Only svs is handled currently. x:{hdr['xcsi']},y:{hdr['ycsi']},z:{hdr['zcsi']}.")

    data,ref_data = _read_data(filename,hdr)

    affine = _calc_affine(hdr)
    orientation = NIFTIOrient(affine)

    dwelltime = 1/hdr['spec_width']

    # Meta - which will be written to the JSON side car requires
    # a few parameters. Set these first. Then set the hdrs from 
    # the files.
    cf = float(hdr["ps_mps_freq"])/1E7
    meta = {'ImagingFrequency':cf,'Dwelltime':dwelltime}   
    meta.update(hdr)   

    return data, ref_data, orientation, dwelltime, meta

def _calc_affine(hdrs):
    warnings.warn("Orientation information for GE files is incomplete. Will appear at iso-centre")

    angles = [0,0,0] # TO DO!

    dimensions = hdrs['vox_size']

    shift = hdrs['vox_offset']

    return calc_affine(angles,dimensions,shift)

def _read_data(filename,hdrs):
    
    frames = int(hdrs['nframes'])
    echoes = int(hdrs['nechoes'])
    fid_pnts = int(hdrs['frame_size'])
    ref_frames = int(hdrs['rhuser19'])

    with open(filename,'rb') as f:
        f.seek(0,2)
        file_size = f.tell()
        f.seek(0)
        pnts = int((file_size-hdrs['off_data'])/4)
        print(pnts)
        f.seek(hdrs['off_data'])
        data = np.fromfile(f,dtype=np.int32,count=pnts)

    coils = 0
    for n in np.arange(0,8,2):
        if hdrs['rcv'][n]!=0 or hdrs['rcv'][n+1]!=0:
            coils = coils + 1 + hdrs['rcv'][n+1] - hdrs['rcv'][n]
    if coils == 0:
        coils = 1
        
    expt_pts = coils * (frames * echoes + echoes) * fid_pnts * 2

    if expt_pts != pnts:
        warnings.warn("Unexpected number of data points.")
        print(f"Expecting {pnts} points based on file size and {expt_pts} based on header.")
        print(f"Coils {coils}")
        print(f"nframes {frames}")
        print(f"nechoes {echoes}")
        print(f"frame_size {fid_pnts}")
        print(f"w_frames {ref_frames}")

    data = data.reshape((-1,2))
    data = data[:,0]+1j*data[:,1]

    dynamics = frames * echoes + echoes
    data = data.reshape((coils,dynamics,fid_pnts))

    data = np.conj(data)

    # Remove empty frame
    data = data[:,1:,]

    data = np.moveaxis(data,1,0)
    if ref_frames == 0:
        ref_data = None
    else:
        ref_data = data[0:ref_frames,:,:].copy()
    data = data[ref_frames:,:,:]

    data = list(data)
    ref_data = list(ref_data)

    return data,ref_data

def _read_hdr(filename):
    with open(filename,'rb') as f:
    
        #RDB version
        f.seek(0)
        rdb_rev = np.fromfile(f,dtype=np.float32,count=1).item()     
        byte_off,interp_off = _rev_info(rdb_rev)
        
        hdr_vars = {}
        f.seek(byte_off['off_data'])
        hdr_vars['off_data'] = np.fromfile(f,dtype=np.int32,count=1).item() 

        f.seek(byte_off['nechoes'])
        hdr_vars['nechoes'] = np.fromfile(f,dtype=np.int16,count=1).item() 
        
        f.seek(byte_off['nframes'])
        hdr_vars['nframes'] = np.fromfile(f,dtype=np.int16,count=1).item()
        
        f.seek(byte_off['frame_size'])
        hdr_vars['frame_size'] = np.fromfile(f,dtype=np.uint16,count=1).item()
        
        f.seek(byte_off['rcv'])
        hdr_vars['rcv'] = np.fromfile(f,dtype=np.int16,count=8)
        
        f.seek(byte_off['rhuser19'])
        hdr_vars['rhuser19'] = np.fromfile(f,dtype=np.float32,count=1).item()
        
        f.seek(byte_off['spec_width'])
        hdr_vars['spec_width'] = np.fromfile(f,dtype=np.float32,count=1).item()
        
        f.seek(byte_off['csi_dims'])
        hdr_vars['csi_dims'] = np.fromfile(f,dtype=np.int16,count=1).item()
        
        f.seek(byte_off['xcsi'])
        hdr_vars['xcsi'] = np.fromfile(f,dtype=np.int16,count=1).item()
        
        f.seek(byte_off['ycsi'])
        hdr_vars['ycsi'] = np.fromfile(f,dtype=np.int16,count=1).item()
        
        f.seek(byte_off['zcsi'])
        hdr_vars['zcsi'] = np.fromfile(f,dtype=np.int16,count=1).item()
        
        f.seek(byte_off['ps_mps_freq'])
        hdr_vars['ps_mps_freq'] = np.fromfile(f,dtype=np.uint32,count=1).astype(float).item()
        
        f.seek(0)
        count = max(interp_off['rdb_hdr_off_image'], interp_off['rdb_hdr_ps_mps_freq']) 
        i_hdr_value = np.fromfile(f,dtype=np.int32,count=count)
        f.seek(i_hdr_value[interp_off['rdb_hdr_off_image']-1])
        t_hdr_value = np.fromfile(f,dtype=np.int32,count=count)    
        
        o_hdr_off = i_hdr_value[interp_off['rdb_hdr_off_image']-1]
        f.seek(o_hdr_off)
        o_hdr_value = np.fromfile(f,dtype=np.float32,count=interp_off['brhc']+3) 

    hdr_vars['tr'] = float(t_hdr_value[interp_off['image_tr']]/1E6)
    hdr_vars['te'] = float(t_hdr_value[interp_off['image_te']]/1E6)
    hdr_vars['vox_size'] = o_hdr_value[interp_off['image_user8']:interp_off['image_user8']+3]
    hdr_vars['vox_offset'] = o_hdr_value[interp_off['image_user11']:interp_off['image_user11']+3]
    hdr_vars['tlhc_RAS'] = o_hdr_value[interp_off['tlhc']:interp_off['tlhc']+3]
    hdr_vars['trhc_RAS'] = o_hdr_value[interp_off['trhc']:interp_off['trhc']+3]
    hdr_vars['brhc_RAS'] = o_hdr_value[interp_off['brhc']:interp_off['brhc']+3]

    return hdr_vars

def _rev_info(rev):
    byte_offsets = {}
    interpreted_offsets = {}
    # This section is from spant
    if rev>25:
        byte_offsets['hdr_rev'] = 0
        byte_offsets['off_data'] = 4
        byte_offsets['nechoes'] = 146
        byte_offsets['nframes'] = 150
        byte_offsets['frame_size'] = 156
        byte_offsets['rcv'] = 264
        byte_offsets['rhuser19'] = 356
        byte_offsets['spec_width'] = 432
        byte_offsets['csi_dims'] = 436
        byte_offsets['xcsi'] = 438
        byte_offsets['ycsi'] = 440
        byte_offsets['zcsi'] = 442
        byte_offsets['ps_mps_freq'] = 488
#         byte_offsets['te'] = 1148
    elif rev>11 and rev<25:
        byte_offsets['hdr_rev'] = 0
        byte_offsets['off_data'] = 1468
        byte_offsets['nechoes'] = 70
        byte_offsets['nframes'] = 74
        byte_offsets['frame_size'] = 80
        byte_offsets['rcv'] = 200
        byte_offsets['rhuser19'] = 292
        byte_offsets['spec_width'] = 368
        byte_offsets['csi_dims'] = 372
        byte_offsets['xcsi'] = 374
        byte_offsets['ycsi'] = 376
        byte_offsets['zcsi'] = 378
        byte_offsets['ps_mps_freq'] = 424
#         byte_offsets['te'] = 1212
    else:
        raise ValueError(f'Revision number {rev} not recognised.')
    
    # This section is from gannet and coroborated from pfile-tools.
    if rev == 14.3:
        interpreted_offsets['rdb_hdr_off_image']   = 377
        interpreted_offsets['rdb_hdr_ps_mps_freq'] = 107
        interpreted_offsets['image_user8']         = 37
        interpreted_offsets['image_user11']        = 40
        interpreted_offsets['tlhc']                = 120
        interpreted_offsets['trhc']                = 123
        interpreted_offsets['brhc']                = 126
        interpreted_offsets['image_te']            = 180
        interpreted_offsets['image_tr']            = 178
    elif rev == 16.0:
        interpreted_offsets['rdb_hdr_off_image']   = 377
        interpreted_offsets['rdb_hdr_ps_mps_freq'] = 107
        interpreted_offsets['image_user8']         = 49
        interpreted_offsets['image_user11']        = 52
        interpreted_offsets['tlhc']                = 132
        interpreted_offsets['trhc']                = 135
        interpreted_offsets['brhc']                = 138
        interpreted_offsets['image_te']            = 192
        interpreted_offsets['image_tr']            = 190
    elif rev in [20.006,20.007,24.0]:
        interpreted_offsets['rdb_hdr_off_image']   = 377
        interpreted_offsets['rdb_hdr_ps_mps_freq'] = 107
        interpreted_offsets['image_user8']         = 97
        interpreted_offsets['image_user11']        = 100
        interpreted_offsets['tlhc']                = 180
        interpreted_offsets['trhc']                = 183
        interpreted_offsets['brhc']                = 186
        interpreted_offsets['image_te']            = 266
        interpreted_offsets['image_tr']            = 264
    elif rev == 26.002:
        interpreted_offsets['rdb_hdr_off_image']   = 11
        interpreted_offsets['rdb_hdr_ps_mps_freq'] = 123
        interpreted_offsets['image_user8']         = 97
        interpreted_offsets['image_user11']        = 100
        interpreted_offsets['tlhc']                = 180
        interpreted_offsets['trhc']                = 183
        interpreted_offsets['brhc']                = 186
        interpreted_offsets['image_te']            = 266
        interpreted_offsets['image_tr']            = 264
    else:
        raise ValueError(f'Revision number {rev} not recognised.')
    
    return byte_offsets,interpreted_offsets




import numpy as np
from spec2nii.dcm2niiOrientation.orientationFuncs import nifti_mat44_to_quatern
from scipy.spatial.transform import Rotation
class NIFTIOrient:
    def __init__(self,affine):        
        self.Q44 = affine   
        qb,qc,qd,qx,qy,qz,dx,dy,dz,qfac = nifti_mat44_to_quatern(affine)
        self.qb = qb
        self.qc = qc
        self.qd = qd
        self.qx = qx
        self.qy = qy
        self.qz = qz
        self.dx = dx
        self.dy = dy
        self.dz = dz
        self.qfac = qfac


def calc_affine(angles,dimensions,shift):

    scalingMat = np.diag(dimensions)
    rot = Rotation.from_euler('xyz', angles, degrees=True)
    m33 = rot.as_matrix()@scalingMat
    m44 = np.zeros((4,4))
    m44[0:3,0:3] = m33
    m44[3,3] = 1.0
    m44[0:3,3] = shift

    return m44


import numpy as np

def fGSLAlmEqual(dArg1, dArg2):
# Python implementation of IDEA-VB17/n4/pkg/MrServers/MrMeasSrv/SeqFW/libGSL/fGSLAlmEqual.cpp
  dTmp = dArg1 - dArg2
  bOut =  (dTmp >= -1.0e-6)  &  (dTmp <= 1.0e-6)
  return bOut

def fGSLClassOri(dSagComp,dCorComp,dTraComp,DEBUG):
    # Python implementation of IDEA-VB17/n4/pkg/MrServers/MrMeasSrv/SeqFW/libGSL/fGSLClassOri.cpp
    # This function determines whether a supplied normal vector
    # *	              describes a sagittal, coronal or transverse slice.
    # *		Result:
    # *		  CASE = 0: Sagittal
    # *		  CASE = 1: Coronal
    # *		  CASE = 2: Transverse
    # *     This was created with the following table:
    # *			(Note: ~ means 'about equal')
    # *
    # *		  |sag|~|cor| and |sag|~|tra|  -->  tra
    # *
    # *		  |sag|~|cor| and |sag|<|tra|  -->  tra
    # *		  |sag|~|cor| and |sag|>|tra|  -->  cor
    # *
    # *		  |sag|~|tra| and |sag|<|cor|  -->  cor
    # *		  |sag|~|tra| and |sag|>|cor|  -->  tra
    # *
    # *		  |cor|~|tra| and |cor|<|sag|  -->  sag
    # *		  |cor|~|tra| and |cor|>|sag|  -->  tra
    # *
    # *		  |sag|>|cor| and |sag|>|tra|  -->  sag
    # *		  |sag|>|cor| and |sag|<|tra|  -->  tra
    # *		  |sag|<|cor| and |cor|>|tra|  -->  cor
    # *		  |sag|<|cor| and |cor|<|tra|  -->  tra
    # *
    # *		  |sag|>|tra| and |sag|<|cor|  -->  cor
    # *		  |sag|>|tra| and |sag|>|cor|  -->  sag
    # *		  |sag|<|tra| and |tra|<|cor|  -->  cor
    # *		  |sag|<|tra| and |tra|>|cor|  -->  tra
    # *
    # *		  |cor|>|tra| and |cor|<|sag|  -->  sag
    # *		  |cor|>|tra| and |cor|>|sag|  -->  cor
    # *		  |cor|<|tra| and |tra|<|sag|  -->  sag
    # *		  |cor|<|tra| and |tra|>|sag|  -->  tra
    # In:
    # double  dSagComp,	   /* IMP: Sagittal component of normal vector          */
    # double  dCorComp,	   /* IMP: Coronal component of normal vector           */
    # double  dTraComp,	   /* IMP: Transverse component of normal vector        */
    # Out:
    # long *  plCase       /* EXP: Case (0=Sagittal, 1=Coronal or 2=Transverse) */

    if DEBUG:
        print(f'Normal vector = {dSagComp : 10.7f} {dCorComp : 10.7f} {dTraComp : 10.7f}.',dSagComp, dCorComp, dTraComp)

    #Compute some temporary values
    dAbsSagComp     = np.abs(dSagComp)
    dAbsCorComp     = np.abs(dCorComp)
    dAbsTraComp     = np.abs(dTraComp)
    
    bAlmEqualSagCor = fGSLAlmEqual(dAbsSagComp, dAbsCorComp)
    bAlmEqualSagTra = fGSLAlmEqual(dAbsSagComp, dAbsTraComp)
    bAlmEqualCorTra = fGSLAlmEqual(dAbsCorComp, dAbsTraComp)
        
    #Check all values to determine the slice orientation (sag, cor, tra)

    if ((bAlmEqualSagCor              &  bAlmEqualSagTra)             |
        (bAlmEqualSagCor              &  (dAbsSagComp < dAbsTraComp)) |
        (bAlmEqualSagTra              &  (dAbsSagComp > dAbsCorComp)) |
        (bAlmEqualCorTra              &  (dAbsCorComp > dAbsSagComp)) |
        ((dAbsSagComp > dAbsCorComp)  &  (dAbsSagComp < dAbsTraComp)) |
        ((dAbsSagComp < dAbsCorComp)  &  (dAbsCorComp < dAbsTraComp)) |
        ((dAbsSagComp < dAbsTraComp)  &  (dAbsTraComp > dAbsCorComp)) |
        ((dAbsCorComp < dAbsTraComp)  &  (dAbsTraComp > dAbsSagComp))):
        #Mainly transverse...
        if DEBUG:
            print('Mainly transverse.')        
        plcase = 2 #CASE = 2: Transverse
        
    elif ((bAlmEqualSagCor              &  (dAbsSagComp > dAbsTraComp)) |
            (bAlmEqualSagTra              &  (dAbsSagComp < dAbsCorComp)) |
            ((dAbsSagComp < dAbsCorComp)  &  (dAbsCorComp > dAbsTraComp)) |
            ((dAbsSagComp > dAbsTraComp)  &  (dAbsSagComp < dAbsCorComp)) |
            ((dAbsSagComp < dAbsTraComp)  &  (dAbsTraComp < dAbsCorComp))): 
            #Mainly coronal...
            if DEBUG:
                print('Mainly coronal.')        
            plcase = 1 # CASE = 1: Coronal
    
    elif ((bAlmEqualCorTra              &  (dAbsCorComp < dAbsSagComp)) |
            ((dAbsSagComp > dAbsCorComp)  &  (dAbsSagComp > dAbsTraComp)) |
            ((dAbsCorComp > dAbsTraComp)  & (dAbsCorComp < dAbsSagComp)) |
            ((dAbsCorComp < dAbsTraComp)  &  (dAbsTraComp < dAbsSagComp))):  
            # Mainly sagittal...
            if DEBUG:
                print('Mainly coronal.')
            plcase = 0 # CASE = 0: Sagital

    else:    #Invalid slice orientation...
        raise ValueError('Error: Invalid slice orientation')    

    return plcase

def fGSLCalcPRS(dGs,dPhi,DEBUG):    
    # Python implementation of IDEA-VB17/n4/pkg/MrServers/MrMeasSrv/SeqFW/libGSL/fGSLCalcPRS.cpp
    # * Description : calculates the two vectors of phase encoding and and
    # *               readout direction.                                 
    # *               The calculation depends on the slice orientation (the slice
    # *               normal vector) and on the angle of rotation around the s axis.
    # *		            Every vector (Gp, Gr and Gs) has the three components sag, cor
    # *               and tra, the description is patient oriented. All three
    # *               vectors have a length of 1.0. The biggest component of Gs must
    # *               have a positive sign.
    # *
    # *		Formulas for the rotation around the vektor s:
    # *		    (a = cos (dPhi), b = sin (dPhi))
    # *
    # *		    new             old               rotation     base
    # *		    vector  =   coordinate system   *  matrix    * vector
    # *
    # *		    (P_sag)   (P_sag  R_sag  S_sag)   ( a  b  0)   (1)
    # *		    (P_cor) = (P_cor  R_cor  S_cor) * (-b  a  0) * (0)
    # *		    (P_tra)   (P_tra  R_tra  S_tra)   ( 0  0  0)   (0)
    # *
    # *		    (R_sag)   (P_sag  R_sag  S_sag)   ( a  b  0)   (0)
    # *		    (R_cor) = (P_cor  R_cor  S_cor) * (-b  a  0) * (1)
    # *		    (R_tra)   (P_tra  R_tra  S_tra)   ( 0  0  0)   (0)
    # *
    # *		    (S_sag)   (P_sag  R_sag  S_sag)   ( a  b  0)   (0)
    # *		    (S_cor) = (P_cor  R_cor  S_cor) * (-b  a  0) * (0)
    # *		    (S_tra)   (P_tra  R_tra  S_tra)   ( 0  0  0)   (1)
    # *
    # *		    This multiplied:
    # *
    # *		    (P_sag)   (a * P_sag - b * R_sag)
    # *		    (P_cor) = (a * P_cor - b * R_cor)
    # *		    (P_tra)   (a * P_tra - b * R_tra)
    # *
    # *		    (R_sag)   (b * P_sag + a * R_sag)
    # *		    (R_cor) = (b * P_cor + a * R_cor)
    # *		    (R_tra)   (b * P_tra + a * R_tra)
    # *
    # *		    (S_sag)   (S_sag)
    # *		    (S_cor) = (S_cor)	well!
    # *		    (S_tra)   (P_tra)                
    # %
    # Input:
    #   dGs: The GS vector (= slice normal vector)
    #   dPhi: The rotational angle around Gs
    # Output: 
    #   dGp: The Gp vector
    #   dGr: The Gr vector
    # From libGSL.h    

    #patient axes                                                          */
    SAGITTAL   = 0
    CORONAL    = 1
    TRANSVERSE = 2

    # Start of function
    lOrientation = 0  # Orientation (SAGITTAL, CORONAL or TRANSVERSE)
    lOrientation = lOrientation + fGSLClassOri(dGs[SAGITTAL], dGs[CORONAL], dGs[TRANSVERSE],DEBUG)
    dGp  = np.zeros((3),dtype = float)

    if lOrientation == TRANSVERSE:
        dGp[0] = 0.0
        dGp[1] = dGs[2] * np.sqrt (1. / (dGs[1] * dGs[1] + dGs[2] * dGs[2]))
        dGp[2] = -dGs[1] * np.sqrt (1. / (dGs[1] * dGs[1] + dGs[2] * dGs[2]))
    elif lOrientation == CORONAL:
        dGp[0] = dGs[1] * np.sqrt (1. / (dGs[0] * dGs[0] + dGs[1] * dGs[1]))
        dGp[1] = -dGs[0] * np.sqrt (1. / (dGs[0] * dGs[0] + dGs[1] * dGs[1]))
        dGp[2] = 0.0
    elif lOrientation == SAGITTAL:
        dGp[0] = -dGs[1] * np.sqrt (1. / (dGs[0] * dGs[0] + dGs[1] * dGs[1]))
        dGp[1] = dGs[0] * np.sqrt (1. / (dGs[0] * dGs[0] + dGs[1] * dGs[1]))
        dGp[2] = 0.0
    else:
        raise ValueError('Invalid slice orientation returned from fGSLClassOri')
    
    #Calculate GR = GS x GP 
    dGr = np.zeros((3),dtype =float)
    dGr[0] = dGs[1] * dGp[2] - dGs[2] * dGp[1]
    dGr[1] = dGs[2] * dGp[0] - dGs[0] * dGp[2]
    dGr[2] = dGs[0] * dGp[1] - dGs[1] * dGp[0]

    if DEBUG:
        print('Before rotation around S:')
        print(f'GP = {dGp[0]:10.7f} {dGp[1]:10.7f} {dGp[2]:10.7f}')
        print(f'GR = {dGr[0]:10.7f} {dGr[1]:10.7f} {dGr[2]:10.7f}')
        print(f'GS = {dGs[0]:10.7f} {dGs[1]:10.7f} {dGs[2]:10.7f}')

    #Rotation
    if dPhi != 0.0:
        #Rotate around the S axis                                             
        if DEBUG:
            tmp = dPhi * 180.0 / np.pi
            print(f'PHI = {tmp:10.7f}')
    
    dGp[0] = np.cos(dPhi) * dGp[0] - np.sin(dPhi) * dGr[0]
    dGp[1] = np.cos(dPhi) * dGp[1] - np.sin(dPhi) * dGr[1]
    dGp[2] = np.cos(dPhi) * dGp[2] - np.sin(dPhi) * dGr[2]

    # Calculate new GR = GS x GP                                           
    dGr[0] = dGs[1] * dGp[2] - dGs[2] * dGp[1]
    dGr[1] = dGs[2] * dGp[0] - dGs[0] * dGp[2]
    dGr[2] = dGs[0] * dGp[1] - dGs[1] * dGp[0]
      
    if DEBUG:
        print('After the Rotation around S:')
        print(f'GP = {dGp[0]:10.7f} {dGp[1]:10.7f} {dGp[2]:10.7f}')
        print(f'GR = {dGr[0]:10.7f} {dGr[1]:10.7f} {dGr[2]:10.7f}')
        print(f'GS = {dGs[0]:10.7f} {dGs[1]:10.7f} {dGs[2]:10.7f}')

    return dGp, dGr